## @class
# Requesters are responsible for providing the data requested via
# a "get" statement in tests.cfg (Monitor::Tester).
package Requester::Requester;

use strict;
use TestWorker;

my $pid = $$;

my %domain_handlers = ();
my %cache = ();

my %domain_warned = ();
my %loaded = ();


## @cmethod new()
# create a requester.
sub new
{
    my ($this, $args ) = @_;
    my  $class = ref($this) || $this;
    my  $self = {};

    bless($self, $class);
    return $self;
}


## @method void register( @domains )
# Register a requester for one or multiple domains. Each Requester
# has to register itself under a unique domain name. Typically,
# a requester will create an instance of itself on load (a singleton)
# and register this instance via register(). This instance will then
# get "get" requests for the registered domains.
#
# @param domains The list of domain names to register.
sub register {
    my( $self, @domains ) = @_;
    
    foreach my $domain (@domains) {
        $domain_handlers{$domain} = $self;
    }
}


## @fn protected void preexit()
# This function is called during program shutdown.
sub preexit {
    %domain_handlers = ();
}


## @fn @ probe_consumes( hashref env, arrayref variables )
# Given a list of variable names and an environment, probe_consumes() will
# try to determine the criticalness/consumes values for the requests that
# it takes to get all those variable's values.
#
# @param env Tester environment
# @param variables A list of variable names.
# @return An array with two values: The first one being the criticalness,
#   the 2nd one the consumer level of the most critical or most consuming
#   requester that has to be called. Returns undef if no requester has to
#   be called at all.
sub probe_consumes {
    my( $env, $variables ) = @_;

    my $domains = _get( $env, $variables, 1 );
    return(undef) unless( %$domains );
    my $criticalness = 0;
    my $consumes = 0;
    while( my( $domain, $vars ) = each %$domains ) {
        my $handler = get_handler( $domain );
	next unless $handler;
	$criticalness = $handler->criticalness() if( $handler->criticalness() > $criticalness );
	$consumes = $handler->consumes() if( $handler->consumes() > $consumes );
    }
    return( $criticalness, $consumes );
}


## @fn hashref get( hashref env, arrayref variables )
# Given a list of variable names and a tester environment requests all
# the variables and returns the result in a hash ref.
#
# @param env Tester environment
# @param variables A list of variable names.
# @return A hash ref.
sub get {
    my( $env, $variables ) = @_;
    return( _get( $env, $variables, 0 ) );
}

## @fn protected hashref _get( hashref env, arrayref variables, $probe_only )
# Implements both the get() and probe_consumes() call.
#
# @param env Tester environment
# @param variables List of variable names
# @param probe_only If set use probe_consumes() semantics, otherwise
#   get() is performed
# @return Hashref compatible with get() if called with probe_only not set,
#   otherwise a hashref with the keys being the names of domains that need
#   a requester to be called.
sub _get {
    my( $env, $variables, $probe_only ) = @_;

    my %result = ();
    my %perdomain = ();
    my $frequency = $env->{"sched_frequency"};
    my $time = $env->{"sched_minute"};
    my $host = $env->{"nodeinstance"}->{"host"};
    foreach my $variable (@$variables) {
        if( defined $cache{"$host $variable $frequency $time"} ) {
	    my $cached = $cache{"$host $variable $frequency $time"};
	    return( $cached ) if( UNIVERSAL::isa( $cached, "CWorker" ) && ! $probe_only );
	    $result{$variable} = $cached;
	}
	else {
	    $variable =~ /^(\@|)([^\.]+)\./;
	    my( $qualifier, $domain, $object ) = ($1, $2, $');
	    if( ! defined $perdomain{$domain} ) {
	        $perdomain{$domain} = [];
	    }
	    push( @{$perdomain{$domain}}, $qualifier.$object );
	}
    }
    return( \%perdomain ) if( $probe_only );
    my $returned = \%result;
    while( my( $domain, $vars ) = each %perdomain ) {
        my $handler = get_handler( $domain );
	return( undef ) unless( $handler );
	my $results = $handler->request( $env, $domain, $vars );
	return(undef) unless( $results );
	if( UNIVERSAL::isa( $results, "CWorker" ) ) {;
 	    foreach my $name (@$vars) {
		set_cache( $env, $domain, $name, $results );
	    }
	    $returned = $results;
	}
	else {
	    while( my( $name, $value ) = each %{$results} ) {
		if( $name =~ /^@/ ) {
		    $name = $&.$domain.".".$';
		}
		else {
		    $name = $domain.".".$name;
		}
		$result{$name} = $value;
		$cache{"$host $name $frequency $time"} = $value;
	    }
	}
    }
    return $returned;
}


## @fn protected void set_cache( hashref env, $domain, $name, hashref value )
# Given a domain, variable name and a value stores the value
# in the cache under the given domain/name.
#
# @param env Tester environment
# @param domain Domain name
# @param name Variable name
# @param value Value associated with domain/name. The values being
#   stored in the cache are required to be hash references.
sub set_cache {
    my( $env, $domain, $name, $value ) = @_;

    if( $name =~ /^@/ ) {
	$name = $&.$domain.".".$';
    }
    else {
	$name = $domain.".".$name;
    }
    my $host = $env->{"nodeinstance"}->{"host"};
    my $frequency = $env->{"sched_frequency"};
    my $time = $env->{"sched_minute"};
    $cache{"$host $name $frequency $time"} = $value;
}



## @method hashref request( hashref env, $domain, arrayref variables )
# Computes (or requests) the values of the variables listed in variables. The
# default request() method will process the list of variables and then call
# getwalk() if the request includes a walk, getinfo() otherwise.
#
# @param env Tester environment
# @param domain The name of the domain the request is associated with. If the
#   requester in question is registered for one single domain only, this is
#   always the name of the registered domain.
# @param variables List of variable names that are requested. Variable names
#   starting with '@' mean that the requester is supposed to return all
#   variables/values below this prefix.
# @return A hashref containing the variable names associated with their values.
#   Returns 'undef' if the requester cannot perform the request (i.e. because
#   the variable names are unknown or the underlying system failed)
sub request {
    my( $self, $env, $domain, $variables ) = @_;

    $self->{"env"} = $env;
    $self->{"domain"} = $domain;
    my @gets;
    my @walks;
    my @indexes;
    foreach my $variable (@$variables) {
        if( $variable =~ /^@/ ) {
	    push( @walks, $' );
	}
	else {
	    push( @gets, $variable );
	    if( $variable =~ /\.(\d+)$/ ) {
	        my $idx = $1;
	        push( @indexes, $idx ) unless( grep( $_ eq $idx, @indexes ) );
	    }
	}
    }

    my $time = time;
    my $info;
    if( @walks ) {
        $info = $self->getwalk( \@gets, \@indexes, \@walks );
    }
    else {
        $info = $self->getinfo( \@gets, \@indexes );
    }
    $time = ($time + time)/2;

    my %result = ();
    my $code = sub {
	foreach my $instance (keys %{$info}) {
	    next unless( ref $info->{$instance} );
	    while( my( $var, $val ) = each %{$info->{$instance}} ) {
		set_cache( $env, $domain, "$var.$instance", { 0 => { 
		    "time" => time,
		    "value" => $val 
		} } );
	    }
	}

	foreach my $get (@gets) {
	    my $val;
	    if( ($get =~ /\.(\d+)$/) && defined $info->{$1}->{$`} ) {
		my $idx = $1;
		my $var = $`;
		$result{$get} = $val = { 0 => {
		    "time" => time,
		    "value" => $info->{$idx}->{$var}
		} };
	    }
	    elsif( $self->fail_novar() ) {
	        return();
	    }
	    else {
		$result{$get} = $val = { };
	    }
	    set_cache( $env, $domain, $get, $val );
	}

	foreach my $walk (@walks) {
	    my $this = {};
	    foreach my $idx (keys %$info) {
		$this->{$idx} = {
		    "time" => $time,
		    "value" => $info->{$idx}->{$walk}
		}
	    }
	    $result{"\@$walk"} = $this;
	    set_cache( $env, $domain, "\@$walk", $this );
	}
	return 1;
    };
    if( UNIVERSAL::isa( $info, "CWorker" ) ) {
	# if getinfo() deferred we just create another CWorker
	# waiting for getinfo() to finish and then storing the
	# results in the cache
        my $worker = new TestWorker();
	$worker->start_after( $info );
	$worker->do_next( sub {
	    $info = $info->{"Test.info"};
	    return 1;
	} );
	$worker->do_next( $code );
	return $worker;
    }
    else {
        $code->();
	return \%result;
    }

}


## @method boolean fail_novar( boolean fail )
# Not used any more?
# 
# @todo check if fail_novar() is still in use by someone, and if yes, what the
#   hell it is used for
sub fail_novar {
    my( $self, $fail ) = @_;

    $self->{"fail_novar"} = $fail if( defined $fail );
    return( $self->{"fail_novar"} );
}


## @method hashref getinfo( arrayref gets, arrayref indexes )
# Called by the default request() method with gets set to the list
# of the variables, indexes set to the list of variable indexes to
# be retrieved. If gets or indexes are undefined, this means that
# a 'walk' through all the possible variables has to be performed.
#
# The default behaviour is to call the getwalk() method with an
# empty walks list. So, a requester implementation has to implement
# at least one of request(), getinfo() or getwalk() in order to do
# something useful.
#
# @param gets The list of names of the variables that have to be
#   retrieved.
# @param indexes The list of variable indexes that have to be
#   retrieved.
# @return A folded hashref, at the first level the hashref uses the
#   variable index as key, the values each point to a hashref with
#   the name of the respective variable as a key, the scalar value as the
#   value.
sub getinfo {
    my( $self, $gets, $indexes ) = @_;
    return( $self->getwalk( $gets, $indexes, [] ) ) if( $gets && $indexes );
    return {};
}


## @method hashref getwalk( arrayref gets, arrayref indexes, arrayref walks )
# Similiar to getinfo(), but getwalk() is called by the default request() method
# if the request includes a "walk" and passes the list of variables to be walked
# via the walks parameter.
#
# The default behaviour of getwalk() is to just call
# getinfo() with gets/indexes unset. This has been the default behaviour before
# getwalk() has been introduced, so getwalk() behaves the same in order to avoid
# breaking existing requesters. A requester implementation has to implement at
# least one of request(), getinfo() or getwalk() in order to do something useful.
#
# @param gets The list of names of the variables that have to be
#   retrieved.
# @param indexes The list of variable indexes that have to be
#   retrieved.
# @param walks The list of variables that have to be "walked"
# @return A folded hashref, at the first level the hashref uses the
#   variable index as key, the values each point to a hashref with
#   the name of the respective variable as a key, the scalar value as the
#   value.
sub getwalk {
    my( $self, $gets, $indexes, $walks ) = @_;
    return $self->getinfo();
}


## @fn protected void clear_cache()
# Clear the variable cache.
sub clear_cache {
    %cache = ();
}


## @fn Requester::Requester get_handler( $domain )
# Given the name of a domain this function searches for a request handler
# that is willing to perform requests for this domain. If no handler has
# been registered yet, it tries to dynamically load one via load_module().
#
# @param domain The name of the domain
# @return The handler of the domain or undef if none was found.
sub get_handler {
    my( $domain ) = @_;

    unless( defined $domain_handlers{$domain} ) {
        load_module( $domain );
    }
    unless( defined $domain_handlers{$domain} ) {
	unless( $domain_warned{$domain} ) {
	    BigSister::common::log( "err", "Requester: no handler for domain $domain" );
	    $domain_warned{$domain} = 1;
	}
	return undef;
    }
    return $domain_handlers{$domain};
}


## @fn boolean load_module( $module )
# Given the name of a module tries to dynamically load Requester::modulename.
#
# @param module The name of a module
# @return true if successful, false otherwise
sub load_module {
    my( $module ) = @_;
    my( $stat, $dir, $file );
    my( $ret );

    return if( $loaded{$module} );
    $ret = 0;
    eval {
	require "Requester/$module.pm";
	$loaded{$module} = 1;
	$ret = 1;
    };
    return $ret;
}


## @method boolean available( hashref env )
# Determine if a requester is willing to perform (handle requests) for a given
# tester environment. Usually, available() will return false if some required system
# command is not installed, some API is missing, the test works only against a local
# machine but the target host is not the local machine, etc.
#
# @param env Tester environment
# @return true if requester is willing to handle requests, false otherwise
sub available {
    my( $self, $env ) = @_;

    return( 1 );
}


## @method boolean command_available( @cmd )
# Tests if cmd is runnable - usually called by classes inherited from Requester::Requester
# in their available() method in order to find out if a system command they require is
# present or not.
#
# @note The command in question is actually executed via system(), so it is required
#   that the cmd argument includes "save" command line options ensuring that no harm
#   is done to the system and Big Sister.
#
# @param cmd The command to be tested for availability.
sub command_available {
    my( $self, @cmd ) = @_;

	return( 0 ) unless( $Platform::has_fork );
	my $pid = open( _SUB, "-|" );
	if( $pid ) {
	    while( <_SUB> ) {;}
		close _SUB;
		return( ! $? );
    }
	else {
		close STDERR;
		open( STDERR, ">&STDOUT" );
	    exit( (system( @cmd ) & 0xFF == 0xFF) ? 1 : 0 );
    }
}


## @method int criticalness( int criticalness )
# Get/Set the criticalness of a handler. The higher the criticalness the more the
# handler will suffer from other handlers to be running in parallel.
#
# @param criticalness (optional) If defined, set the criticalness to this value
# @return The level of criticalness.
sub criticalness {
    my( $self, $criticalness ) = @_;

    $self->{"criticalness"} = $criticalness if( defined $criticalness );
    return( $self->{"criticalness"} || 2 );
}


## @method int consumes( int consumes )
# Get/set consume level of a handler. The higher the "consumes" figure is the
# more other handlers will suffer if running in parallel with this handler.
# 
# @param consumes (optional) if defined, set the consumes figure to this value
# @return The current 'consumes' figure.
sub consumes {
    my( $self, $consumes ) = @_;

    $self->{"consumes"} = $consumes if( defined $consumes );
    return( defined( $self->{"consumes"} ) ? $self->{"consumes"} : 3 );
}



END {
    if( $pid == $$ ) {
	    preexit();
    }
}

1;
